import 'rxjs/add/operator/toPromise';
import { AsyncSubject as AsyncSubject$1 } from 'rxjs/AsyncSubject';
import { Observable as Observable$1 } from 'rxjs/Observable';
import 'rxjs/add/operator/distinctUntilChanged';
import { Inject, Injectable, InjectionToken, Injector, NgModule, Optional, Pipe, isDevMode } from '@angular/core';
import 'rxjs/add/operator/map';
import { CommonModule } from '@angular/common';
import 'rxjs/add/operator/takeWhile';
import { BehaviorSubject as BehaviorSubject$1 } from 'rxjs/BehaviorSubject';
import { createStore } from 'redux';

class ActionInterface {
}

class MetadataManager {
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setActionMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getActionMetadata(target) {
        return MetadataManager.get(target, MetadataManager.ACTION_KEY, MetadataManager.ACTION_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setActionContextMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_CONTEXT_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getActionContextMetadata(target) {
        return MetadataManager.get(target, MetadataManager.ACTION_CONTEXT_KEY, MetadataManager.ACTION_CONTEXT_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setReducerMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.REDUCER_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getReducerMetadata(target) {
        return MetadataManager.get(target, MetadataManager.REDUCER_KEY, MetadataManager.REDUCER_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setStateMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.STATE_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getStateMetadata(target) {
        return MetadataManager.get(target, MetadataManager.STATE_KEY, MetadataManager.STATE_DEFAULT);
    }
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} data
     * @return {?}
     */
    static set(target, key, data) {
        Reflect['defineMetadata'](key, data, target);
    }
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} defaultData
     * @return {?}
     */
    static get(target, key, defaultData) {
        defaultData = defaultData || ({});
        return Object.assign({}, defaultData, Reflect['getMetadata'](key, target));
    }
}
MetadataManager.ACTION_KEY = Symbol('@ReduxAction');
MetadataManager.ACTION_DEFAULT = {
    contextClazz: null,
    type: '',
};
MetadataManager.ACTION_CONTEXT_KEY = Symbol('@ReduxActionContext');
MetadataManager.ACTION_CONTEXT_DEFAULT = {
    prefix: '',
};
MetadataManager.REDUCER_KEY = Symbol('@ReduxReducer');
MetadataManager.REDUCER_DEFAULT = {
    reducers: [],
};
MetadataManager.STATE_KEY = Symbol('@ReduxState');
MetadataManager.STATE_DEFAULT = {
    name: '',
};

/**
 * @param {?} config
 * @return {?}
 */
function ReduxActionContext(config) {
    return (constructor) => {
        MetadataManager.setActionContextMetadata(constructor, config);
        return constructor;
    };
}

class Registry {
    /**
     * @return {?}
     */
    static reset() {
        Registry._store = new AsyncSubject$1();
        Registry._reducers = [];
    }
    /**
     * @param {?} store
     * @return {?}
     */
    static registerStore(store) {
        Registry._store.next(store);
        Registry._store.complete();
    }
    /**
     * @param {?} stateName
     * @param {?} types
     * @param {?} reducer
     * @return {?}
     */
    static registerReducer(stateName, types, reducer) {
        if (Array.isArray(types)) {
            types.forEach((type) => {
                Registry._reducers.push({
                    reducer,
                    stateName,
                    type,
                });
            });
        }
        else {
            Registry._reducers.push({
                reducer,
                stateName,
                type: /** @type {?} */ (types),
            });
        }
    }
    /**
     * @param {?} state
     * @return {?}
     */
    static registerState(state) {
        Registry.getStore().then((store) => {
            const /** @type {?} */ stateConfig = MetadataManager.getStateMetadata(state.constructor);
            const /** @type {?} */ initState = state.getInitialState();
            let /** @type {?} */ initStateToResolve = initState;
            if (initState instanceof Observable$1) {
                initStateToResolve = initState.toPromise();
            }
            Promise.resolve(initStateToResolve).then((initialState) => {
                store.dispatch({
                    payload: {
                        initialValue: initialState,
                        name: stateConfig.name,
                    },
                    type: Registry.ACTION_REGISTER_STATE,
                });
            });
        });
    }
    /**
     * @param {?} type
     * @return {?}
     */
    static getReducerItemsByType(type) {
        return Registry._reducers
            .filter(reducerItem => getActionTypeByFunction(reducerItem.type) === type);
    }
    /**
     * @return {?}
     */
    static getStore() {
        return new Promise(Registry._store.subscribe.bind(Registry._store));
    }
}
Registry.ACTION_REGISTER_STATE = `@harmowatch/ngx-redux-core/registerState`;
Registry._store = new AsyncSubject$1();
Registry._reducers = [];

/**
 * @param {?=} config
 * @return {?}
 */
function ReduxAction(config) {
    return (target, propertyKey, descriptor) => {
        config = Object.assign({
            type: propertyKey,
        }, config || {});
        const /** @type {?} */ originalFunction = descriptor.value;
        const /** @type {?} */ proxyFunction = function () {
            const /** @type {?} */ contextData = MetadataManager.getActionContextMetadata(target.constructor);
            const /** @type {?} */ type = getActionType(contextData.prefix, config.type);
            const /** @type {?} */ returnValue = originalFunction.apply(this, arguments);
            Promise.all([
                Promise.resolve(returnValue),
                Registry.getStore(),
            ]).then(([payload, store]) => {
                store.dispatch({ payload, type });
            });
            return returnValue;
        };
        MetadataManager.setActionMetadata(proxyFunction, {
            contextClazz: target.constructor,
            type: config.type,
        });
        descriptor.value = proxyFunction;
    };
}
/**
 * @param {?} target
 * @return {?}
 */
function getActionTypeByFunction(target) {
    const { type, contextClazz } = MetadataManager.getActionMetadata(target);
    const { prefix } = MetadataManager.getActionContextMetadata(contextClazz);
    return getActionType(prefix, type);
}
/**
 * @param {...?} rest
 * @return {?}
 */
function getActionType(...rest) {
    return rest.filter((i) => i.length > 0).join('');
}

class ReduxModuleRootReducer {
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduce(state, action) {
        if (action.type === Registry.ACTION_REGISTER_STATE) {
            return ReduxModuleRootReducer.reduceRegisterState(state, /** @type {?} */ (action));
        }
        return ReduxModuleRootReducer.reduceByRegisteredReducers(state, action);
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduceRegisterState(state, action) {
        const { name, initialValue } = action.payload;
        return Object.assign({}, state, {
            [name]: initialValue,
        });
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduceByRegisteredReducers(state, action) {
        const /** @type {?} */ reducers = Registry.getReducerItemsByType(action.type);
        return reducers.reduce((stateToReduce, reducerItem) => {
            const { stateName, reducer } = reducerItem;
            return Object.assign({}, stateToReduce, {
                [stateName]: reducer(stateToReduce[stateName], action),
            });
        }, state);
    }
}

/**
 * @template S, P
 * @param {?} types
 * @return {?}
 */
function ReduxReducer(types) {
    return (target, propertyKey) => {
        const /** @type {?} */ data = MetadataManager.getReducerMetadata(target.constructor);
        data.reducers.push({ types, reducer: target[propertyKey].bind(target) });
        MetadataManager.setReducerMetadata(target.constructor, data);
        return target[propertyKey];
    };
}

class ReduxSelector extends Observable$1 {
    /**
     * @param {?=} selector
     * @param {?=} stateProvider
     */
    constructor(selector = '/', stateProvider) {
        super(observer => {
            Registry.getStore().then(store => {
                store.subscribe(() => {
                    observer.next(ReduxSelector.getValueByState(store.getState(), selector, stateProvider));
                });
            });
        });
        if (!ReduxSelectorCacheFactory.has(selector, stateProvider)) {
            ReduxSelectorCacheFactory.set(this, selector, stateProvider);
        }
    }
    /**
     * @template T
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static create(selector, stateProvider) {
        return new ReduxSelector(selector, stateProvider);
    }
    /**
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static getAbsoluteSelector(selector, stateProvider) {
        if (!selector.startsWith(ReduxSelector.DELIMITER)) {
            if (!stateProvider) {
                throw new Error('You need to provide a state provider, if you use relative selectors');
            }
            return `/${MetadataManager.getStateMetadata(stateProvider).name}/${selector}`;
        }
        return selector;
    }
    /**
     * @template S
     * @param {?} state
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static getValueByState(state, selector, stateProvider) {
        return ReduxSelector.getAbsoluteSelector(selector, stateProvider).split(ReduxSelector.DELIMITER)
            .filter(propertyKey => propertyKey !== '')
            .reduce((previousValue, propertyKey) => {
            if (!previousValue || !previousValue.hasOwnProperty(propertyKey)) {
                return null;
            }
            return previousValue[propertyKey];
        }, /** @type {?} */ (state));
    }
}
ReduxSelector.DELIMITER = '/';

class ReduxSelectorCacheFactory {
    /**
     * @return {?}
     */
    static clear() {
        ReduxSelectorCacheFactory.selectors = {};
    }
    /**
     * @template T
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static getOrCreate(selector, stateProvider) {
        if (ReduxSelectorCacheFactory.has(selector, stateProvider)) {
            return ReduxSelectorCacheFactory.get(selector, stateProvider);
        }
        return new ReduxSelector(selector, stateProvider);
    }
    /**
     * @template T
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static get(selector, stateProvider) {
        const /** @type {?} */ key = ReduxSelectorCacheFactory.getKey(selector, stateProvider);
        return (ReduxSelectorCacheFactory.selectors[key]);
    }
    /**
     * @param {?} reduxSelector
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static set(reduxSelector, selector, stateProvider) {
        const /** @type {?} */ key = ReduxSelectorCacheFactory.getKey(selector, stateProvider);
        ReduxSelectorCacheFactory.selectors[key] = reduxSelector;
    }
    /**
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static has(selector, stateProvider) {
        const /** @type {?} */ key = ReduxSelectorCacheFactory.getKey(selector, stateProvider);
        return !!ReduxSelectorCacheFactory.selectors[key];
    }
    /**
     * @param {?} selector
     * @param {?=} stateProvider
     * @return {?}
     */
    static getKey(selector, stateProvider) {
        return ReduxSelector.getAbsoluteSelector(selector, stateProvider);
    }
}
ReduxSelectorCacheFactory.selectors = {};

class StateDefToken extends InjectionToken {
    constructor() {
        super('StateDefToken');
    }
}

class ReduxSelectPipe {
    /**
     * @param {?=} stateDefs
     */
    constructor(stateDefs = []) {
        this.stateDef = stateDefs[0];
    }
    /**
     * @param {?} selector
     * @return {?}
     */
    transform(selector) {
        return ReduxSelectorCacheFactory.getOrCreate(selector, this.stateDef.provider)
            .distinctUntilChanged();
    }
}
ReduxSelectPipe.decorators = [
    { type: Pipe, args: [{ name: 'reduxSelect' },] },
];
/**
 * @nocollapse
 */
ReduxSelectPipe.ctorParameters = () => [
    { type: Array, decorators: [{ type: Inject, args: [StateDefToken,] },] },
];

/**
 * @template S
 * @param {?} expression
 * @param {?=} context
 * @return {?}
 */
function ReduxSelect(expression, context) {
    return (target, propertyKey) => {
        let /** @type {?} */ observable;
        Object.defineProperty(target, propertyKey, {
            configurable: true,
            enumerable: true,
            get: () => {
                if (!observable) {
                    observable = new ReduxSelector(expression, context).distinctUntilChanged();
                }
                return observable;
            },
        });
    };
}

/**
 * @deprecated Use "ReduxSelector" instead
 */
class ReduxStateSelector {
    /**
     * @param {?} expression
     * @param {?=} context
     */
    constructor(expression, context) {
        this.selector = new ReduxSelector(expression, context);
    }
    /**
     * @return {?}
     */
    asObservable() {
        return this.selector;
    }
}

let ReduxStateSelectorSubjectType = {};
ReduxStateSelectorSubjectType.DEFAULT = 0;
ReduxStateSelectorSubjectType.SUBJECT = 1;
ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT = 2;
ReduxStateSelectorSubjectType.REPLAY_SUBJECT = 3;
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.DEFAULT] = "DEFAULT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.SUBJECT] = "SUBJECT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT] = "BEHAVIOR_SUBJECT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.REPLAY_SUBJECT] = "REPLAY_SUBJECT";

/**
 * @param {?} config
 * @return {?}
 */
function ReduxState(config) {
    return (constructor) => {
        MetadataManager.setStateMetadata(constructor, config);
        return constructor;
    };
}

class ReduxStore extends InjectionToken {
    constructor() {
        super('ReduxStore');
    }
}

class ReduxCommonModule {
}
ReduxCommonModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxCommonModule.ctorParameters = () => [];

class ReduxTestingStore {
    constructor() {
        this.state = new BehaviorSubject$1(null);
    }
    /**
     * @template S
     * @param {?} state
     * @param {?} value
     * @return {?}
     */
    setState(state, value) {
        const { name } = MetadataManager.getStateMetadata(state);
        const /** @type {?} */ nextState = Object.assign({}, this.state.getValue(), {
            [name]: value,
        });
        this.state.next(nextState);
        return this.state
            .takeWhile(currentState => currentState !== nextState)
            .toPromise().then(() => this.state.getValue());
    }
    /**
     * @return {?}
     */
    getState() {
        return this.state.getValue();
    }
    /**
     * @param {?} listener
     * @return {?}
     */
    subscribe(listener) {
        return this.state.subscribe.call(this.state, listener);
    }
    /**
     * @return {?}
     */
    replaceReducer() {
    }
    /**
     * @template T
     * @param {?} action
     * @return {?}
     */
    dispatch(action) {
        return action;
    }
}
ReduxTestingStore.decorators = [
    { type: Injectable },
];
/**
 * @nocollapse
 */
ReduxTestingStore.ctorParameters = () => [];

class ReduxTestingModule {
    /**
     * @param {?=} store
     */
    constructor(store = null) {
        ReduxSelectorCacheFactory.clear();
        if (store) {
            Registry.reset();
            Registry.registerStore(store);
        }
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    static forRoot(config = {}) {
        return {
            ngModule: ReduxTestingModule,
            providers: config.state ? [
                { provide: ReduxTestingStore, useFactory: config.storeFactory || ReduxTestingModule.defaultStoreFactory },
                { provide: StateDefToken, useValue: config.state || null, multi: true },
                config.state ? config.state.provider : null,
            ] : [
                { provide: ReduxTestingStore, useFactory: config.storeFactory || ReduxTestingModule.defaultStoreFactory },
                { provide: StateDefToken, useValue: config.state || null, multi: true },
            ],
        };
    }
    /**
     * @return {?}
     */
    static defaultStoreFactory() {
        return new ReduxTestingStore();
    }
}
ReduxTestingModule.decorators = [
    { type: NgModule, args: [{
                exports: [
                    ReduxCommonModule,
                ],
                imports: [
                    CommonModule,
                ],
                providers: [
                    ReduxTestingStore,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxTestingModule.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [ReduxTestingStore,] },] },
];

class StateDefinitionManager {
    /**
     * @param {?} stateDef
     * @return {?}
     */
    static registerReducers(stateDef) {
        if (stateDef.reducers) {
            const /** @type {?} */ stateMetadata = MetadataManager.getStateMetadata(stateDef.provider);
            stateDef.reducers
                .map((reducer) => MetadataManager.getReducerMetadata(reducer.constructor))
                .forEach((metadata) => {
                metadata.reducers.forEach(reducer => {
                    Registry.registerReducer(stateMetadata.name, reducer.types, reducer.reducer);
                });
            });
        }
    }
}

class ReduxModule {
    /**
     * @param {?=} stateDefs
     * @param {?=} store
     * @param {?=} injector
     */
    constructor(stateDefs = [], store = null, injector) {
        this.injector = injector;
        if (store) {
            Registry.registerStore(store);
        }
        stateDefs
            .filter(stateDef => !!stateDef)
            .map(stateDef => {
            return {
                metadata: MetadataManager.getStateMetadata(stateDef.provider),
                stateDef,
            };
        })
            .filter(({ metadata }) => !ReduxModule.knownStateDefinitions[metadata.name])
            .forEach(({ metadata, stateDef }) => {
            this.initState(stateDef);
            ReduxModule.knownStateDefinitions[metadata.name] = stateDef;
        });
    }
    /**
     * @param {?} stateDef
     * @return {?}
     */
    initState(stateDef) {
        Registry.registerState(this.injector.get(stateDef.provider));
        StateDefinitionManager.registerReducers(stateDef);
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    static forChild(config = {}) {
        return {
            ngModule: ReduxModule,
            providers: [
                { provide: StateDefToken, useValue: config.state || null, multi: true },
                config.state ? config.state.provider : null,
            ],
        };
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    static forRoot(config = {}) {
        return {
            ngModule: ReduxModule,
            providers: config.state ? [
                { provide: ReduxStore, useFactory: config.storeFactory || ReduxModule.defaultStoreFactory },
                { provide: StateDefToken, useValue: config.state || null, multi: true },
                config.state ? config.state.provider : null,
            ] : [
                { provide: ReduxStore, useFactory: config.storeFactory || ReduxModule.defaultStoreFactory },
                { provide: StateDefToken, useValue: config.state || null, multi: true },
            ],
        };
    }
    /**
     * @return {?}
     */
    static defaultStoreFactory() {
        return createStore(ReduxModuleRootReducer.reduce, {}, ReduxModule.defaultEnhancerFactory());
    }
    /**
     * @return {?}
     */
    static defaultEnhancerFactory() {
        if (console && window['__REDUX_DEVTOOLS_EXTENSION__'] && isDevMode()) {
            return window['__REDUX_DEVTOOLS_EXTENSION__']();
        }
        return ReduxModule.nullEnhancer;
    }
    /**
     * @param {?} next
     * @return {?}
     */
    static nullEnhancer(next) {
        return next;
    }
}
ReduxModule.knownStateDefinitions = {};
ReduxModule.decorators = [
    { type: NgModule, args: [{
                exports: [
                    ReduxCommonModule,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxModule.ctorParameters = () => [
    { type: Array, decorators: [{ type: Optional }, { type: Inject, args: [StateDefToken,] },] },
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [ReduxStore,] },] },
    { type: Injector, },
];

/**
 * Generated bundle index. Do not edit.
 */

export { ActionInterface, ReduxActionContext, ReduxAction, getActionTypeByFunction, getActionType, ReduxModuleRootReducer, ReduxReducer, ReduxSelect, ReduxSelectPipe, ReduxStateSelector, ReduxStateSelectorSubjectType, ReduxState, ReduxStore, ReduxTestingModule, ReduxTestingStore, ReduxModule, ReduxCommonModule as ɵb, StateDefToken as ɵa };
//# sourceMappingURL=ngx-redux-core.js.map
